// newton distance estimator (GPLv3+) 2013 Claude Heiland-Allen <claude@mathr.co.uk>
// gcc -std=c99 -O3 -Wall -pedantic -Wextra -o newt newt.c -lm && ./newt

#include <complex.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

typedef unsigned int N;
typedef double R;
typedef double complex C;

R distance(C z0, N nzero, const C *zero, const C *zerop, N npole, const C *pole, const C *polep, N *which) {
  C z = z0;
  C dz = 1.0;
  R eps = 0.1;
  for (N k = 0; k < 1024; ++k) {
    for (N i = 0; i < nzero; ++i) {
      R e = cabs(z - zero[i]);
      if (e < eps) {
        *which = i;
        return e * -log(e) / cabs(dz);
      }
    }
    C sz = 0.0;
    C sz2 = 0.0;
    for (N i = 0; i < nzero; ++i) {
      C d = z - zero[i];
      sz += zerop[i] / d;
      sz2 += zerop[i] / (d * d);
    }
    C sp = 0.0;
    C sp2 = 0.0;
    for (N j = 0; j < npole; ++j) {
      C d = z - pole[j];
      sp += polep[j] / d;
      sp2 += polep[j] / (d * d);
    }
    C d = sz - sp;
    z -= 1.0 / d;
    dz *= (sp2 - sz2) / (d * d) + 1.0;
  }
  *which = nzero;
  return -1;
}

int main(int argc, char **argv) {
  if (argc != 6) {
    fprintf(stderr, "usage: %s [width] [height] [real] [imag] [size] > out.ppm\n", argv[0]);
    return 1;
  }
  N width = atoi(argv[1]);
  N height = atoi(argv[2]);
  C z0 = atof(argv[3]) + atof(argv[4]) * I;
  R r = atof(argv[5]);
  R pixel_size = r * 2.0 / height;

  // {{{ customize {{{
  N nzero = 4;
  C zero[4] = { 1, -I, -1, I };
  C zerop[4] = { 1, 1, 1, 1 };
  C npole = 2;
  C pole[2] = { -2, 2 };
  C polep[2] = { 1, 1 };
  R rgb[5][3] = { {1,0.7,0.7}, {1,1,0.7}, {0.7,1,0.7}, {0.7,0.7,1}, {1,1,1} };
  // }}} customize }}}

  printf("P6\n%u %u\n255\n", width, height);
  for (N y = 0; y < height; ++y) {
    for (N x = 0; x < width; ++x) {
      R py = height/2.0 - (y + 0.5);
      R px = (x + 0.5) - width/2.0;
      C z = z0 + (px + py * I) * pixel_size;
      N w = 0;
      R d = distance(z, nzero, zero, zerop, npole, pole, polep, &w);
      R y = fmin(fmax(tanh(d / pixel_size) * 256, 0), 255);
      int red = y * rgb[w][0];
      int grn = y * rgb[w][1];
      int blu = y * rgb[w][2];
      putchar(red);
      putchar(grn);
      putchar(blu);
    }
  }
  return 0;
}
